/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#include "lislib.h"
#include "lis_io.h"


int lis_input_mm(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm, int matrix_type, ...)
{
	int				bnr,bnc;
	int				err;
	int				*crow,*ccol;
	va_list			vvlist;

	va_start( vvlist, matrix_type );
	bnr = va_arg(vvlist, int);
	bnc = va_arg(vvlist, int);
	crow = va_arg(vvlist, int *);
	ccol = va_arg(vvlist, int *);
	va_end(vvlist);

	switch( matrix_type )
	{
	case LIS_MATRIX_CRS:
		err = lis_input_mm_crs(A,b,x,file,comm,matrix_type,bnr,bnc,crow,ccol);
		break;
	case LIS_MATRIX_CCS:
		err = lis_input_mm_ccs(A,b,x,file,comm);
		break;
	case LIS_MATRIX_MSR:
		err = lis_input_mm_msr(A,b,x,file,comm);
		break;
	case LIS_MATRIX_DIA:
		err = lis_input_mm_dia(A,b,x,file,comm);
		break;
	case LIS_MATRIX_ELL:
		err = lis_input_mm_ell(A,b,x,file,comm);
		break;
	case LIS_MATRIX_JDS:
		err = lis_input_mm_jds(A,b,x,file,comm);
		break;
	case LIS_MATRIX_BSR:
		err = lis_input_mm_crs(A,b,x,file,comm,matrix_type,bnr,bnc,crow,ccol);
		break;
	case LIS_MATRIX_BSC:
		err = lis_input_mm_crs(A,b,x,file,comm,matrix_type,bnr,bnc,crow,ccol);
		break;
	case LIS_MATRIX_VBR:
		err = lis_input_mm_crs(A,b,x,file,comm,matrix_type,bnr,bnc,crow,ccol);
		break;
	case LIS_MATRIX_DNS:
		err = lis_input_mm_dns(A,b,x,file,comm);
		break;
	case LIS_MATRIX_COO:
		err = lis_input_mm_coo(A,b,x,file,comm);
		break;
	}
	return LIS_SUCCESS;
}

int lis_input_mm_vec(LIS_MATRIX A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm, int isb, int isx)
{
	char			buf[BUFSIZE];
	int				i,k,j;
	int				err;
	int				n,is,ie;
	int				idx;
	LIS_SCALAR		val;

	if( isb==0 && isx==0 ) return LIS_SUCCESS;

	n   = A->gn;
	is  = A->is;
	ie  = A->ie;
	if( isb )
	{
		lis_vector_duplicateM(A,b);
		for(i=0;i<n;i++)
		{
			if( fgets(buf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( sscanf(buf, "%d %lg", &idx, &val) != 2 )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			idx--;
			if( idx>=is && idx<ie )
			{
				(*b)->value[idx-is] = val;
			}
		}
	}
	if( isx )
	{
		lis_vector_duplicateM(A,x);
		for(i=0;i<n;i++)
		{
			if( fgets(buf, BUFSIZE, file) == NULL )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			if( sscanf(buf, "%d %lg", &idx, &val) != 2 )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			idx--;
			if( idx>=is && idx<ie )
			{
				(*x)->value[idx-is] = val;
			}
		}
	}
	return LIS_SUCCESS;
}

int lis_input_mm_crs(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm, int matrix_type,int bnr, int bnc, int *row, int *col)
{
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,local_nnz;
	int				i,k,j,my_rank;
	int				err;
	int				mmtype;
	int				n,is,ie;
	int				ridx,cidx;
	int				*ptr, *index;
	int				*work;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n = Ain->n;
	lis_matrix_get_range(Ain,&is,&ie);
	ptr   = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	work  = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( work==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	for(i=0;i<n+1;i++) ptr[i]=0;
	for(i=0;i<n+1;i++) work[i]=0;
	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file)==NULL )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>is && cidx<=ie ) work[cidx-is-1]++;
			}
			if( ridx>is && ridx<=ie )
			{
				ptr[ridx-is]++;
			}
		}
	}

	ptr[0] = 0;
	for( i=0; i<n; i++ )
	{
		if( mmtype==MM_SYMM )
		{
			ptr[i+1] += ptr[i] + work[i];
		}
		else
		{
			ptr[i+1] += ptr[i];
		}
		work[i] = 0;
	}

	index   = (int *)lis_malloc( ptr[n]*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value   = (LIS_SCALAR *)lis_malloc( ptr[n]*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	rewind(file);
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>=is && cidx<ie )
				{
					value[ptr[cidx-is]+work[cidx-is]] = val;
					index[ptr[cidx-is]+work[cidx-is]] = ridx;
					work[cidx-is]++;
				}
			}
			if( ridx>=is && ridx<ie )
			{
				value[ptr[ridx-is]+work[ridx-is]] = val;
				index[ptr[ridx-is]+work[ridx-is]] = cidx;
				work[ridx-is]++;
			}
		}
	}

	err = lis_matrix_set_crs(ptr[n],ptr,index,value,Ain);
	if( err )
	{
		return err;
	}

	err = lis_matrix_assemble(&Ain,matrix_type);
	if( err ) return err;

	if( matrix_type==LIS_MATRIX_CRS )
	{
		*A = Ain;
	}
	else
	{
		err = lis_matrix_convert(Ain,A,matrix_type,bnr,bnc,row,col);
		if( err ) return err;
	}

	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		lis_free(work);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	lis_free(work);

	return LIS_SUCCESS;
}

int lis_input_mm_ccs(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,local_nnz;
	int				i,k,j,my_rank;
	int				err;
	int				mmtype;
	int				n,is,ie;
	int				ridx,cidx;
	int				*ptr, *index, *row, *col;
	int				*work;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n = Ain->n;
	lis_matrix_get_range(Ain,&is,&ie);
	ptr   = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	work  = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( work==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	for(i=0;i<n+1;i++) ptr[i]=0;
	for(i=0;i<n+1;i++) work[i]=0;
	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file)==NULL )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( ridx>is && ridx<=ie ) work[ridx-is-1]++;
			}
			if( cidx>is && cidx<=ie )
			{
				ptr[cidx-is]++;
			}
		}
	}

	ptr[0] = 0;
	for( i=0; i<n; i++ )
	{
		if( mmtype==MM_SYMM )
		{
			ptr[i+1] += ptr[i] + work[i];
		}
		else
		{
			ptr[i+1] += ptr[i];
		}
		work[i] = 0;
	}

	index   = (int *)lis_malloc( ptr[n]*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value   = (LIS_SCALAR *)lis_malloc( ptr[n]*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	rewind(file);
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( ridx>=is && ridx<ie )
				{
					value[ptr[ridx-is]+work[ridx-is]] = val;
					index[ptr[ridx-is]+work[ridx-is]] = cidx;
					work[ridx-is]++;
				}
			}
			if( cidx>=is && cidx<ie )
			{
				value[ptr[cidx-is]+work[cidx-is]] = val;
				index[ptr[cidx-is]+work[cidx-is]] = ridx;
				work[cidx-is]++;
			}
		}
	}

	err = lis_matrix_set_ccs(ptr[n],ptr,index,value,Ain);
	if( err )
	{
		return err;
	}

	err = lis_matrix_assemble(&Ain,LIS_MATRIX_CCS);
	if( err ) return err;

	*A = Ain;
	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		lis_free(work);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	lis_free(work);

	return LIS_SUCCESS;
}

int lis_input_mm_msr(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,ndz,local_nnz;
	int				i,k,j,my_rank;
	int				err;
	int				mmtype;
	int				n,is,ie;
	int				ridx,cidx;
	int				*index;
	int				*work;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n = Ain->n;
	lis_matrix_get_range(Ain,&is,&ie);
	index   = (int *)lis_malloc( (nnz+1)*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	work  = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( work==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	for(i=0;i<n+1;i++) index[i]=0;
	for(i=0;i<n+1;i++) work[i]=0;
	ndz = n;
	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file)==NULL )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>is && cidx<=ie ) work[cidx-is-1]++;
			}
			if( ridx>is && ridx<=ie )
			{
				if( ridx!=cidx )
				{
					index[ridx-is]++;
				}
				else
				{
					ndz--;
				}
			}
		}
	}

	index[0] = n+1;
	for( i=0; i<n; i++ )
	{
		if( mmtype==MM_SYMM )
		{
			index[i+1] += index[i] + work[i];
		}
		else
		{
			index[i+1] += index[i];
		}
		work[i] = 0;
	}

	if( (nnz+1)!=index[n] )
	{
		index   = (int *)realloc( index, index[n]*sizeof(int) );
		if( index==NULL )
		{
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
	}
	value   = (LIS_SCALAR *)lis_malloc( index[n]*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	rewind(file);
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>=is && cidx<ie )
				{
					value[index[cidx-is]+work[cidx-is]] = val;
					index[index[cidx-is]+work[cidx-is]] = ridx;
					work[cidx-is]++;
				}
			}
			if( ridx>=is && ridx<ie )
			{
				if( ridx!=cidx )
				{
					value[index[ridx-is]+work[ridx-is]] = val;
					index[index[ridx-is]+work[ridx-is]] = cidx;
					work[ridx-is]++;
				}
				else
				{
					value[ridx-is] = val;
				}
			}
		}
	}

	err = lis_matrix_set_msr(index[n]-ndz-1,ndz,index,value,Ain);
	if( err )
	{
		return err;
	}

	err = lis_matrix_assemble(&Ain,LIS_MATRIX_MSR);
	if( err ) return err;

	*A = Ain;
	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		lis_free(work);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	lis_free(work);

	return LIS_SUCCESS;
}

int lis_input_mm_dia(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm)
{
#ifdef _OPENMP
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,nnd,local_nnz;
	int				i,k,j,jj,maxj;
	int				err;
	int				mmtype;
	int				n,gn,is,ie,np,nprocs,my_rank;
	int				ridx,cidx;
	int				*index,*col;
	int				*work;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n      = Ain->n;
	gn     = Ain->gn;
	np     = 0;
	nprocs = omp_get_max_threads();
	lis_matrix_get_range(Ain,&is,&ie);

	work  = (int *)lis_malloc( nprocs*(n+gn)*sizeof(int) );
	if( work==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	col   = (int *)lis_malloc( nprocs*sizeof(int) );
	if( col==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

#pragma omp parallel private(i,j,is,ie,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
	for(i=0;i<n+gn;i++) work[my_rank*(n+gn) + i] = 0;
}

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file)==NULL )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
#pragma omp parallel private(i,j,k,is,ie,my_rank)
{
		my_rank = omp_get_thread_num();
		LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
		k       = my_rank*(n+gn);

		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>is && cidx<=ie )
				{
					j = ridx-1;
					work[k + n + j - (cidx-1)] = 1;
				}
			}
			if( ridx>is && ridx<=ie )
			{
				j = cidx-1;
				work[k + n + j - (ridx-1)] =1;
			}
		}
}
	}

#pragma omp parallel private(i,j,k,is,ie,my_rank)
{
		my_rank = omp_get_thread_num();
		LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
		k       = my_rank*(n+gn);

		col[my_rank] = 0;
		for(i=0;i<(n+gn);i++)
		{
			if( work[k + i] > 0 ) col[my_rank]++;
		}
}
	nnd = 0;
	for(i=0;i<nprocs;i++)
	{
		if( nnd < col[i] ) nnd = col[i];
	}

	index   = (int *)lis_malloc( nprocs*nnd*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value   = (LIS_SCALAR *)lis_malloc( n*nnd*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	rewind(file);
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );

#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
	k       = my_rank*(n+gn);

	memset(&value[is*nnd],0,(ie-is)*nnd*sizeof(LIS_SCALAR));
	j = 0;
	for( i=0; i<n+gn && j<col[my_rank]; i++ )
	{
		if( work[k + i]>0 )
		{
			index[my_rank*nnd + j] = i-n;
			work[k + i]            = is*nnd + j*(ie-is);
			j++;
		}
	}
}
	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( val!=0.0 )
		{
#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
			my_rank = omp_get_thread_num();
			LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
			k       = my_rank*(n+gn);

			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>=is && cidx<ie )
				{
					value[work[k + n + ridx - cidx] + cidx-is] = val;
				}
			}
			if( ridx>=is && ridx<ie )
			{
				value[work[k + n + cidx - ridx] + ridx-is] = val;
			}
}
		}
	}

	err = lis_matrix_set_dia(col,index,value,Ain);
	if( err )
	{
		return err;
	}

	err = lis_matrix_assemble(&Ain,LIS_MATRIX_DIA);
	if( err ) return err;

	*A = Ain;
	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		lis_free(work);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	lis_free(work);

	return LIS_SUCCESS;
#else
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,nnd,local_nnz;
	int				i,k,j,jj,maxj,my_rank;
	int				err;
	int				mmtype;
	int				n,gn,is,ie,np;
	int				ridx,cidx;
	int				*index;
	int				*work;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n  = Ain->n;
	gn = Ain->gn;
	np = 0;
	lis_matrix_get_range(Ain,&is,&ie);

	work  = (int *)lis_malloc( (n+gn)*sizeof(int) );
	if( work==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	for(i=0;i<n+gn;i++) work[i] = 0;

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file)==NULL )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>is && cidx<=ie )
				{
					j = ridx-1;
					work[n + j - (cidx-is-1)] = 1;
				}
			}
			if( ridx>is && ridx<=ie )
			{
				j = cidx-1;
				work[n + j - (ridx-is-1)] =1;
			}
		}
	}

	nnd = 0;
	for( i=0; i<n+gn; i++ )
	{
		if( work[i]>0 )
		{
			nnd++;
		}
	}

	index   = (int *)lis_malloc( nnd*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value   = (LIS_SCALAR *)lis_malloc( n*nnd*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	memset(value,0,n*nnd*sizeof(LIS_SCALAR));
	j = 0;
	for( i=0; i<n+gn && j<nnd; i++ )
	{
		if( work[i]>0 )
		{
			index[j] = i-is-n;
			work[i]  = j*n;
			j++;
		}
	}

	rewind(file);
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>=is && cidx<ie )
				{
					value[work[n + ridx - (cidx-is)] + cidx-is] = val;
				}
			}
			if( ridx>=is && ridx<ie )
			{
				value[work[n + cidx - (ridx-is)] + ridx-is] = val;
			}
		}
	}

	err = lis_matrix_set_dia(nnd,index,value,Ain);
	if( err )
	{
		return err;
	}

	err = lis_matrix_assemble(&Ain,LIS_MATRIX_DIA);
	if( err ) return err;

	*A = Ain;
	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		lis_free(work);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	lis_free(work);

	return LIS_SUCCESS;
#endif
}

int lis_input_mm_ell(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,maxnzr,local_nnz;
	int				i,k,j,my_rank;
	int				err;
	int				mmtype;
	int				n,is,ie;
	int				ridx,cidx;
	int				*index;
	int				*work;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n = Ain->n;
	lis_matrix_get_range(Ain,&is,&ie);

	work  = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( work==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	for(i=0;i<n+1;i++) work[i]=0;
	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file)==NULL )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>is && cidx<=ie ) work[cidx-is-1]++;
			}
			if( ridx>is && ridx<=ie )
			{
				work[ridx-is-1]++;
			}
		}
	}

	maxnzr = 0;
	for( i=0; i<n; i++ )
	{
		if( maxnzr<work[i] ) maxnzr = work[i];
		work[i] = i;
	}

	index   = (int *)lis_malloc( n*maxnzr*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value   = (LIS_SCALAR *)lis_malloc( n*maxnzr*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	memset(index,0,n*maxnzr*sizeof(int));
	memset(value,0,n*maxnzr*sizeof(LIS_SCALAR));

	rewind(file);
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>=is && cidx<ie )
				{
					value[work[cidx-is]] = val;
					index[work[cidx-is]] = ridx;
					work[cidx-is]       += n;
				}
			}
			if( ridx>=is && ridx<ie )
			{
				value[work[ridx-is]] = val;
				index[work[ridx-is]] = cidx;
				work[ridx-is]       += n;
			}
		}
	}

	err = lis_matrix_set_ell(maxnzr,index,value,Ain);
	if( err )
	{
		return err;
	}

	err = lis_matrix_assemble(&Ain,LIS_MATRIX_ELL);
	if( err ) return err;

	*A = Ain;
	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		lis_free(work);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	lis_free(work);

	return LIS_SUCCESS;
}

int lis_input_mm_jds(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm)
{
#ifdef _OPENMP
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,nnz2,maxnzr,local_nnz,nprocs,my_rank;
	int				i,k,j;
	int				err;
	int				mmtype;
	int				n,is,ie;
	int				ridx,cidx;
	int				*col,*ptr,*row,*index;
	int				*work,*iw,*iw2;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n      = Ain->n;
	nprocs = omp_get_max_threads();
	lis_matrix_get_range(Ain,&is,&ie);

	row   = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	col   = (int *)lis_malloc( nprocs*sizeof(int) );
	if( col==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	work  = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( work==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	iw  = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	iw2  = (int *)lis_malloc( n*sizeof(int) );
	if( iw2==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

#pragma omp parallel private(i,is,ie,my_rank,maxnzr)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is;i<ie;i++)
	{
		work[i]=0;
	}
}
	work[n] = 0;
	nnz2 = 0;
	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file)==NULL )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( val!=0.0 )
		{
#pragma omp parallel private(i,is,ie,my_rank,maxnzr)
{
			my_rank = omp_get_thread_num();
			LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>is && cidx<=ie )
				{
					work[cidx-1]++;
					nnz2++;
				}
			}
			if( ridx>is && ridx<=ie )
			{
				work[ridx-1]++;
				nnz2++;
			}
}
		}
	}

#pragma omp parallel private(i,is,ie,my_rank,maxnzr)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is;i<ie;i++)
	{
		row[i]=i;
	}
	lis_sortr_ii(is,ie-1,work,row);

	maxnzr = 0;
	for(i=is;i<ie;i++)
	{
		if( maxnzr<work[i] ) maxnzr = work[i];
		iw2[i] = ie-is;
	}
	col[my_rank] = maxnzr;
}
	maxnzr = 0;
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr<col[i] ) maxnzr = col[i];
	}

	ptr   = (int *)lis_malloc( nprocs*(maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	iw[0] = 0;
#pragma omp parallel private(i,j,k,is,ie,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=0;i<col[my_rank]+1;i++)
	{
		ptr[my_rank*(maxnzr+1) + i] = 0;
	}
	k = 0;
	for(i=is;i<ie;i++)
	{
		for(j=0;j<work[i];j++)
		{
			ptr[my_rank*(maxnzr+1) + j+1]++;
		}
		k += work[i];
	}
	iw[my_rank+1] = k;
#pragma omp barrier
#pragma omp single
	for(i=0;i<nprocs;i++)
	{
		iw[i+1] += iw[i];
	}
	ptr[my_rank*(maxnzr+1)] = iw[my_rank];
	for(i=0;i<col[my_rank];i++)
	{
		ptr[my_rank*(maxnzr+1) + i+1] += ptr[my_rank*(maxnzr+1) + i];
	}
	for(i=is;i<ie;i++)
	{
		work[row[i]] = i;
		row[i]       = 0;
	}
}

	index   = (int *)lis_malloc( nnz2*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value   = (LIS_SCALAR *)lis_malloc( nnz2*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	rewind(file);
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( val!=0.0 )
		{
#pragma omp parallel private(i,j,k,is,ie,my_rank)
{
			my_rank = omp_get_thread_num();
			LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>=is && cidx<ie )
				{
					k        = ptr[my_rank*(maxnzr+1) + row[cidx]] + work[cidx]-is;
					value[k] = val;
					index[k] = ridx;
					row[cidx]++;
				}
			}
			if( ridx>=is && ridx<ie )
			{
				k        = ptr[my_rank*(maxnzr+1) + row[ridx]] + work[ridx]-is;
				value[k] = val;
				index[k] = cidx;
				row[ridx]++;
			}
}
		}
	}

#pragma omp parallel private(i,j,k,is,ie,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
	for(i=is;i<ie;i++)
	{
		row[work[i]] = i;
	}
}
	err = lis_matrix_set_jds(nnz2,col,row,ptr,index,value,Ain);
	if( err )
	{
		return err;
	}

	err = lis_matrix_assemble(&Ain,LIS_MATRIX_JDS);
	if( err ) return err;

	*A = Ain;
	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		lis_free2(3,work,iw,iw2);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	lis_free2(3,work,iw,iw2);

	return LIS_SUCCESS;
#else
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,maxnzr,local_nnz;
	int				i,k,j,my_rank;
	int				err;
	int				mmtype;
	int				n,is,ie;
	int				ridx,cidx;
	int				*ptr,*row,*index;
	int				*work;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n = Ain->n;
	lis_matrix_get_range(Ain,&is,&ie);

	row   = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	work  = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( work==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	for(i=0;i<n+1;i++) work[i]=0;
	for(i=0;i<n;i++) row[i]=i;
	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file)==NULL )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>is && cidx<=ie ) work[cidx-is-1]++;
			}
			if( ridx>is && ridx<=ie )
			{
				work[ridx-is-1]++;
			}
		}
	}

	lis_sortr_ii(0,n-1,work,row);
	maxnzr = 0;
	for( i=0; i<n; i++ )
	{
		if( maxnzr<work[i] ) maxnzr = work[i];
	}

	ptr   = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	for(i=0;i<maxnzr+1;i++) ptr[i] = 0;
	for(i=0;i<n;i++)
	{
		for(j=0;j<work[i];j++)
		{
			ptr[j+1]++;
		}
	}
	for(i=0;i<n;i++)
	{
		work[row[i]] = i;
		row[i]       = 0;
	}
	ptr[0] = 0;
	for(i=0;i<maxnzr;i++)
	{
		ptr[i+1] += ptr[i];
	}

	index   = (int *)lis_malloc( ptr[maxnzr]*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value   = (LIS_SCALAR *)lis_malloc( ptr[maxnzr]*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	rewind(file);
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>=is && cidx<ie )
				{
					j        = cidx-is;
					k        = ptr[row[j]] + work[j];
					value[k] = val;
					index[k] = ridx;
					row[j]++;
				}
			}
			if( ridx>=is && ridx<ie )
			{
				j        = ridx-is;
				k        = ptr[row[j]] + work[j];
				value[k] = val;
				index[k] = cidx;
				row[j]++;
			}
		}
	}

	for(i=0;i<n;i++)
	{
		row[work[i]] = i+is;
	}
	err = lis_matrix_set_jds(ptr[maxnzr],maxnzr,row,ptr,index,value,Ain);
	if( err )
	{
		return err;
	}

	err = lis_matrix_assemble(&Ain,LIS_MATRIX_JDS);
	if( err ) return err;

	*A = Ain;
	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		lis_free(work);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	lis_free(work);

	return LIS_SUCCESS;
#endif
}


int lis_input_mm_dns(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,local_nnz;
	int				i,k,j,my_rank;
	int				err;
	int				mmtype;
	int				n,is,ie;
	int				ridx,cidx;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n = Ain->n;
	lis_matrix_get_range(Ain,&is,&ie);

	value   = (LIS_SCALAR *)lis_malloc( nr*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	memset(value,0,nr*n*sizeof(LIS_SCALAR));

	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( mmtype==MM_SYMM && ridx!=cidx )
		{
			if( cidx>=is && cidx<ie )
			{
				value[(ridx-is)*n + cidx] = val;
			}
		}
		if( ridx>=is && ridx<ie )
		{
			value[cidx*n + ridx-is] = val;
		}
	}

	err = lis_matrix_set_dns(value,Ain);
	if( err )
	{
		return err;
	}
	err = lis_matrix_assemble(&Ain,LIS_MATRIX_DNS);
	if( err )
	{
		return err;
	}
	*A = Ain;
	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	return LIS_SUCCESS;
}

int lis_input_mm_coo(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				nr,nc,nnz,local_nnz;
	int				i,k,j,my_rank;
	int				err;
	int				mmtype;
	int				n,is,ie;
	int				ridx,cidx;
	int				*row, *col;
	int				isb,isx;
	LIS_SCALAR		val;
	LIS_SCALAR		*value;
	LIS_MATRIX		Ain;

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif
	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_MTX, strlen(MM_MTX))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))==0 )
	{
		mmtype = MM_GENERAL;
	}
	else if( strncmp(dstruct, MM_TYPE_SYMM, strlen(MM_TYPE_SYMM))==0)
	{
		mmtype = MM_SYMM;
	}
	else
	{
		SETERR("Not general or symmetric\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	err = sscanf(buf, "%d %d %d %d %d", &nr, &nc, &nnz, &isb, &isx);
	if( err==3 )
	{
		isb = 0;
		isx = 0;
	}

	if( my_rank==0 ) printf("%d x %d matrix  %d entries\n",nr,nc,nnz);
	if( nr!=nc )
	{
		SETERR("matrix is not square\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_matrix_create(0,nr,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n = Ain->n;
	lis_matrix_get_range(Ain,&is,&ie);

	local_nnz = 0;
	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file)==NULL )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file I/O error\n");
			return LIS_FAILS;
		}
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>is && cidx<=ie ) local_nnz++;
			}
			if( ridx>is && ridx<=ie )
			{
				local_nnz++;
			}
		}
	}

	row   = (int *)lis_malloc( local_nnz*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	col   = (int *)lis_malloc( local_nnz*sizeof(int) );
	if( col==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value   = (LIS_SCALAR *)lis_malloc( local_nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	rewind(file);
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );

	k = 0;
	for( i=0; i<nnz; i++ )
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %d %lg", &ridx, &cidx, &val) != 3 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		ridx--;
		cidx--;
		if( val!=0.0 )
		{
			if( mmtype==MM_SYMM && ridx!=cidx )
			{
				if( cidx>=is && cidx<ie )
				{
					value[k] = val;
					row[k]   = cidx;
					col[k]   = ridx;
					k++;
				}
			}
			if( ridx>=is && ridx<ie )
			{
				value[k] = val;
				row[k]   = ridx;
				col[k]   = cidx;
				k++;
			}
		}
	}

	err = lis_matrix_set_coo(local_nnz,row,col,value,Ain);
	if( err )
	{
		return err;
	}
	err = lis_matrix_assemble(&Ain,LIS_MATRIX_COO);
	if( err )
	{
		return err;
	}
	*A = Ain;
	err = lis_input_mm_vec(*A,b,x,file,comm,isb,isx);
	if( err )
	{
		lis_matrix_destroy(*A);
		*A = NULL;
		*b = NULL;
		*x = NULL;
	}
	return LIS_SUCCESS;
}
