/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include "lislib.h"
#include "lis_matrix.h"

int lis_matrix_get_diagonal(LIS_MATRIX A, LIS_VECTOR D)
{
	LIS_SCALAR *d;

	d = D->value;
	switch( A->matrix_type )
	{
	case LIS_MATRIX_DNS:
		lis_matrix_get_diagonal_dns(A, d);
		break;
	case LIS_MATRIX_CRS:
		lis_matrix_get_diagonal_crs(A, d);
		break;
	case LIS_MATRIX_CCS:
		lis_matrix_get_diagonal_ccs(A, d);
		break;
	case LIS_MATRIX_MSR:
		lis_matrix_get_diagonal_msr(A, d);
		break;
	case LIS_MATRIX_DIA:
		lis_matrix_get_diagonal_dia(A, d);
		break;
	case LIS_MATRIX_ELL:
		lis_matrix_get_diagonal_ell(A, d);
		break;
	case LIS_MATRIX_JDS:
		lis_matrix_get_diagonal_jds(A, d);
		break;
	case LIS_MATRIX_BSR:
		lis_matrix_get_diagonal_bsr(A, d);
		break;
	case LIS_MATRIX_BSC:
		lis_matrix_get_diagonal_bsc(A, d);
		break;
	case LIS_MATRIX_VBR:
		lis_matrix_get_diagonal_vbr(A, d);
		break;
	case LIS_MATRIX_COO:
		lis_matrix_get_diagonal_coo(A, d);
		break;
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_dns(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i;
	int n;

	n = A->n;
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		d[i] = A->value[i*n + i];
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_crs(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,is,ie;
	int n;

	n    = A->n;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		d[i] = (LIS_SCALAR)0.0;
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				d[i] = A->value[j];
				break;
			}
		}
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_ccs(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;
	int base;

	n    = A->n;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		d[i] = (LIS_SCALAR)0.0;
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				d[i] = A->value[j];
				break;
			}
		}
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_msr(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;

	n    = A->n;
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		d[i] = A->value[i];
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_dia(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n,nnd;

	n    = A->n;
	nnd  = A->nnd;
	for(j=0;j<nnd;j++)
	{
		if( A->index[j]==0 ) break;
	}
//#pragma omp parallel for private(i)
	for(i=0;i<n;i++)
	{
		d[i] = A->value[j*n+i];
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_ell(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;
	k      = n;
#pragma omp parallel for private(i)
	for(i=0;i<n;i++)
	{
		d[i] = (LIS_SCALAR)0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			if( i==A->index[j*n+i] && d[i]==(LIS_SCALAR)0.0 )
			{
				d[i] = A->value[j*n+i];
				k--;
				if( k==0 ) return LIS_SUCCESS;
			}
		}
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_jds(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k,l;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;
	k      = n;
	for(j=0;j<maxnzr;j++)
	{
		l = 0;
		for(i=A->ptr[j];i<A->ptr[j+1];i++)
		{
			if( A->row[l]==A->index[i] )
			{
				d[A->row[l]] = A->value[i];
				k--;
				if( k==0 ) return LIS_SUCCESS;
			}
			l++;
		}
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_bsr(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k,bi,bj,bjj,nr,nc;
	int bnr,bnc,bs;
	int n;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;
	i   = 0;
	for(bi=0;bi<nr;bi++)
	{
		k = 0;
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			bjj = A->bindex[bj];
			if( i>=bjj*bnc && i<(bjj+1)*bnc )
			{
				for(j=i%bnc;j<bnc&&k<bnr&&i<n;j++)
				{
					d[i] = A->value[bj*bs + j*bnr + k];
					i++;
					k++;
				}
			}
			if( k==bnr ) break;
		}
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_bsc(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k,bi,bj,bii,nr,nc;
	int bnr,bnc,bs;
	int n;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;
	i   = 0;
	for(bj=0;bj<nc;bj++)
	{
		for(bi=A->bptr[bj];bi<A->bptr[bj+1];bi++)
		{
			bii = A->bindex[bi];
			if( i>=bii*bnr && i<(bii+1)*bnr )
			{
				k = i%bnr;
				for(j=i%bnc;j<bnc&&k<bnr&&i<n;j++)
				{
					d[i] = A->value[bi*bs + j*bnr + k];
					i++;
					k++;
				}
			}
			if( i==(bj+1)*bnc ) break;
		}
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_vbr(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k,bi,bj,bjj,nr,nc;
	int bnr,bnc,bs;
	int n;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	i   = 0;
	for(bi=0;bi<nr;bi++)
	{
		k = 0;
		bnr = A->row[bi+1] - A->row[bi];
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			bjj = A->bindex[bj];
			bnc = A->col[bjj+1] - A->col[bjj];
			if( i>=bjj*bnc && i<(bjj+1)*bnc )
			{
				for(j=i%bnc;j<bnc&&k<bnr&&i<n;j++)
				{
					d[i] = A->value[A->ptr[bj] + j*bnr + k];
					i++;
					k++;
				}
			}
			if( k==bnr ) break;
		}
	}
	return LIS_SUCCESS;
}

int lis_matrix_get_diagonal_coo(LIS_MATRIX A, LIS_SCALAR d[])
{
	int k;
	int nnz;

	nnz = A->nnz;
	for(k=0; k<nnz; k++)
	{
		if( A->row[k]==A->col[k] )
		{
			d[A->row[k]] = A->value[k];
		}
	}
	return LIS_SUCCESS;
}
