/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "lislib.h"
#include "lis_matrix.h"

int lis_matrix_scaling(LIS_MATRIX A, LIS_VECTOR B, LIS_VECTOR D, int action)
{
	int			i,n,np;
	LIS_SCALAR  *b,*d;

	n  = A->n;
	np = A->np;
	b  = B->value;
	d  = D->value;

	lis_matrix_get_diagonal(A,D);
	if( action==LIS_SCALE_SYMM_DIAG )
	{
#ifdef USE_MPI
		if( A->np>D->np )
		{
			D->value = (LIS_SCALAR *)lis_realloc(D->value,A->np*sizeof(LIS_SCALAR));
			if( D->value==NULL )
			{
				SETERR("out of memory\n");
				return LIS_OUT_OF_MEMORY;
			}
			d = D->value;
		}
		lis_send_recv(A->commtable,d);
#endif
#pragma omp parallel for private(i)
		for(i=0; i<np; i++)
		{
			d[i] = 1.0 / sqrt(fabs(d[i]));
		}

		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			lis_matrix_scaling_symm_crs(A, d);
			break;
		case LIS_MATRIX_CCS:
			lis_matrix_scaling_symm_ccs(A, d);
			break;
		case LIS_MATRIX_MSR:
			lis_matrix_scaling_symm_msr(A, d);
			break;
		case LIS_MATRIX_DIA:
			lis_matrix_scaling_symm_dia(A, d);
			break;
		case LIS_MATRIX_ELL:
			lis_matrix_scaling_symm_ell(A, d);
			break;
		case LIS_MATRIX_JDS:
			lis_matrix_scaling_symm_jds(A, d);
			break;
		case LIS_MATRIX_BSR:
			lis_matrix_scaling_symm_bsr(A, d);
			break;
		case LIS_MATRIX_BSC:
			lis_matrix_scaling_symm_bsc(A, d);
			break;
		case LIS_MATRIX_VBR:
			lis_matrix_scaling_symm_vbr(A, d);
			break;
		case LIS_MATRIX_DNS:
			lis_matrix_scaling_symm_dns(A, d);
			break;
		case LIS_MATRIX_COO:
			lis_matrix_scaling_symm_coo(A, d);
			break;
		}
	}
	else
	{
#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			d[i] = 1.0 / d[i];
		}
		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			lis_matrix_scaling_crs(A, d);
			break;
		case LIS_MATRIX_CCS:
			lis_matrix_scaling_ccs(A, d);
			break;
		case LIS_MATRIX_MSR:
			lis_matrix_scaling_msr(A, d);
			break;
		case LIS_MATRIX_DIA:
			lis_matrix_scaling_dia(A, d);
			break;
		case LIS_MATRIX_ELL:
			lis_matrix_scaling_ell(A, d);
			break;
		case LIS_MATRIX_JDS:
			lis_matrix_scaling_jds(A, d);
			break;
		case LIS_MATRIX_BSR:
			lis_matrix_scaling_bsr(A, d);
			break;
		case LIS_MATRIX_BSC:
			lis_matrix_scaling_bsc(A, d);
			break;
		case LIS_MATRIX_VBR:
			lis_matrix_scaling_vbr(A, d);
			break;
		case LIS_MATRIX_DNS:
			lis_matrix_scaling_dns(A, d);
			break;
		case LIS_MATRIX_COO:
			lis_matrix_scaling_coo(A, d);
			break;
		}
	}

#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		b[i] = b[i]*d[i];
	}
	return LIS_SUCCESS;
}


void lis_matrix_scaling_dns(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;

	n = A->n;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		for(j=0;j<n;j++)
		{
			A->value[i*n + j] *= d[i];
		}
	}
}

void lis_matrix_scaling_crs(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;

	n    = A->n;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			A->value[j] *= d[i];
		}
	}
}

void lis_matrix_scaling_ccs(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;

	n    = A->n;
#pragma omp parallel for private(i,j)
	for(j=0; j<n; j++)
	{
		for(i=A->ptr[j];i<A->ptr[j+1];i++)
		{
			A->value[i] *= d[A->index[i]];
		}
	}
}

void lis_matrix_scaling_msr(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;

	n    = A->n;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			A->value[j] *= d[i];
		}
	}
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		A->value[i] = (LIS_SCALAR)1.0;
	}
}

void lis_matrix_scaling_dia(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,is,ie;
	int n,nnd;

	n    = A->n;
	nnd  = A->nnd;
	for(j=0;j<nnd;j++)
	{
		if( A->index[j] < 0 )
		{
			is = -A->index[j];
			ie = n;
		}
		else
		{
			is = 0;
			ie = n-A->index[j];
		}
		for(i=is;i<ie;i++)
		{
			A->value[j*n + i] *= d[i];
		}
	}
}

void lis_matrix_scaling_ell(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;

	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			A->value[j*n + i] *= d[i];
		}
	}
}

void lis_matrix_scaling_jds(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;

	for(j=0;j<maxnzr;j++)
	{
		k = 0;
		for(i=A->ptr[j];i<A->ptr[j+1];i++)
		{
			A->value[i] *= d[A->row[k]];
			k++;
		}
	}
}

void lis_matrix_scaling_bsr(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int bi,bj;
	int nr,nc;
	int bnr,bnc;
	int n;

	n    = A->n;
	bnr  = A->bnr;
	bnc  = A->bnc;
	nr   = A->nr;
	nc   = A->nc;

	k = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					A->value[k] *= d[bi*bnr+i];
					k++;
				}
			}
		}
	}
}

void lis_matrix_scaling_bsc(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int bi,bj,bii;
	int nr,nc;
	int bnr,bnc;
	int n;

	n    = A->n;
	bnr  = A->bnr;
	bnc  = A->bnc;
	nr   = A->nr;
	nc   = A->nc;

	k = 0;
	for(bj=0;bj<nc;bi++)
	{
		for(bi=A->bptr[bj];bi<A->bptr[bj+1];bi++)
		{
			bii = A->bindex[bi];
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					A->value[k] *= d[bii*bnr+i];
					k++;
				}
			}
		}
	}
}

void lis_matrix_scaling_vbr(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int bi,bj;
	int nr,nc;
	int n;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;

	k = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			for(j=A->col[A->bindex[bj]];j<A->col[A->bindex[bj]+1];j++)
			{
				for(i=A->row[bi];i<A->row[bi+1];i++)
				{
					A->value[k] *= d[i];
					k++;
				}
			}
		}
	}
}

void lis_matrix_scaling_coo(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int nnz;

	nnz = A->nnz;
	for(j=0; j<nnz; j++)
	{
		i = A->row[j];
		A->value[j] *= d[i];
	}
}

void lis_matrix_scaling_symm_dns(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;

	n = A->n;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		for(j=0;j<n;j++)
		{
			A->value[i*n + j] = A->value[i*n + j]*d[i]*d[j];
		}
	}
}

void lis_matrix_scaling_symm_crs(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;

	n    = A->n;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			A->value[j] = A->value[j]*d[i]*d[A->index[j]];
		}
	}
}

void lis_matrix_scaling_symm_ccs(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;

	n    = A->n;
#pragma omp parallel for private(i,j)
	for(j=0; j<n; j++)
	{
		for(i=A->ptr[j];i<A->ptr[j+1];i++)
		{
			A->value[i] = A->value[i]*d[A->index[i]]*d[j];
		}
	}
}

void lis_matrix_scaling_symm_msr(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n;

	n    = A->n;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			A->value[j] = A->value[j]*d[i]*d[A->index[j]];
		}
	}
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		A->value[i] = A->value[i]*d[i]*d[i];
	}
}

void lis_matrix_scaling_symm_dia(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,is,ie;
	int n,nnd;

	n    = A->n;
	nnd  = A->nnd;
	for(j=0;j<nnd;j++)
	{
		if( A->index[j] < 0 )
		{
			is = -A->index[j];
			ie = n;
		}
		else
		{
			is = 0;
			ie = n-A->index[j];
		}
		for(i=is;i<ie;i++)
		{
			A->value[j*n + i] = A->value[j*n + i]*d[i]*d[i-A->index[j]];
		}
	}
}

void lis_matrix_scaling_symm_ell(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;

	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			A->value[j*n + i] = A->value[j*n + i]*d[i]*d[A->index[j*n + i]];
		}
	}
}

void lis_matrix_scaling_symm_jds(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;

	for(j=0;j<maxnzr;j++)
	{
		k = 0;
		for(i=A->ptr[j];i<A->ptr[j+1];i++)
		{
			A->value[i] = A->value[i]*d[A->row[k]]*d[A->index[i]];
			k++;
		}
	}
}

void lis_matrix_scaling_symm_bsr(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int bi,bj,bjj;
	int nr,nc;
	int bnr,bnc;
	int n;

	n    = A->n;
	bnr  = A->bnr;
	bnc  = A->bnc;
	nr   = A->nr;
	nc   = A->nc;

	k = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			bjj = A->bindex[bj];
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					A->value[k] = A->value[k]*d[bi*bnr+i]*d[bjj*bnc+j];
					k++;
				}
			}
		}
	}
}

void lis_matrix_scaling_symm_bsc(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int bi,bj,bii;
	int nr,nc;
	int bnr,bnc;
	int n;

	n    = A->n;
	bnr  = A->bnr;
	bnc  = A->bnc;
	nr   = A->nr;
	nc   = A->nc;

	k = 0;
	for(bj=0;bj<nc;bi++)
	{
		for(bi=A->bptr[bj];bi<A->bptr[bj+1];bi++)
		{
			bii = A->bindex[bi];
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					A->value[k] = A->value[k]*d[bii*bnr+i]*d[bj*bnc+j];
					k++;
				}
			}
		}
	}
}

void lis_matrix_scaling_symm_vbr(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int bi,bj;
	int nr,nc;
	int n;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;

	k = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			for(j=A->col[A->bindex[bj]];j<A->col[A->bindex[bj]+1];j++)
			{
				for(i=A->row[bi];i<A->row[bi+1];i++)
				{
					A->value[k] = A->value[k]*d[i]*d[j];
					k++;
				}
			}
		}
	}
}

void lis_matrix_scaling_symm_coo(LIS_MATRIX A, LIS_SCALAR d[])
{
	int i,j,k;
	int nnz;

	nnz = A->nnz;
	for(k=0; k<nnz; k++)
	{
		i = A->row[k];
		j = A->col[k];
		A->value[k] = A->value[k]*d[i]*d[j];
	}
}
