/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"
#if defined(USE_SSE2)
#include <emmintrin.h>

int lis_matrix_copy_crs_sse2(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nnz2,is,nprocs,my_rank;
	int			*ptr,*index,*row;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	nnz2    = nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	row     = NULL;

	for(i=0;i<n;i++)
	{
		if( Ain->ptr[i+1]-Ain->ptr[i]%2 ) nnz2++;
	}

	row = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(4,ptr,index,value,row);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,row);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz2*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,row);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz2*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,row);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,row);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* copy crs */
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = Ain->ptr[i+1]-Ain->ptr[i];
		row[i+1] = ptr[i+1] + ptr[i];
		if( ptr[i+1]%2 )
		{
			ptr[i+1] = row[i+1]+1;
		}
		else
		{
			ptr[i+1] = row[i+1];
		}
	}
	for(i=0;i<n;i++)
	{
		jj = ptr[i];
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[jj]   = Ain->value[j];
			index[jj]   = Ain->index[j];
			jj++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	(*Aout)->row = row;
	if( err )
	{
		lis_free2(4,ptr,index,value,row);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
}

void lis_matvec_crs_u2_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2;
	__m128d		x1,x2;
	__m128d		y1,y2;
	__m128d		v1,v2;
	LIS_SCALAR	t[2];
	LIS_SCALAR	*vv;

	n    = A->n;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->row[i+1];
		jj = &A->index[js];
		vv = &A->value[js];
		y1  = _mm_xor_pd(y1,y1);
		for(j=js+1;j<je;j+=2)
		{
			j1  = *(jj++); j2  = *(jj++);
			x1  = _mm_set_pd(x[j2],x[j1]);
			v1  = _mm_load_pd(vv);
			x1  = _mm_mul_pd(x1,v1);
			y1  = _mm_add_pd(y1,x1);
			vv += 2;
		}
		_mm_store_pd(t,y1);
		for(j-=1;j<je;j++)
		{
			j1    = *(jj++);
			t[0] += *(vv++) * x[j1];
		}
		y[i] = t[0]+t[1];
	}
}

void lis_matvec_crs_u4_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4;
	__m128d		x1,x2;
	__m128d		y1,y2;
	__m128d		v1,v2;
	LIS_SCALAR	t[2];
	LIS_SCALAR	*vv;

	n    = A->n;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->row[i+1];
		jj = &A->index[js];
		vv = &A->value[js];
		y1  = _mm_xor_pd(y1,y1);
		y2  = _mm_xor_pd(y2,y2);
		for(j=js+3;j<je;j+=4)
		{
			j1  = *(jj++); j2  = *(jj++);
			x1  = _mm_set_pd(x[j2],x[j1]);
			v1  = _mm_load_pd(vv);
			x1  = _mm_mul_pd(x1,v1);
			y1  = _mm_add_pd(y1,x1);
			vv += 2;

			j3  = *(jj++); j4  = *(jj++);
			x2  = _mm_set_pd(x[j4],x[j3]);
			v2  = _mm_load_pd(vv);
			x2  = _mm_mul_pd(x2,v2);
			y2  = _mm_add_pd(y2,x2);
			vv += 2;
		}
		y1  = _mm_add_pd(y1,y2);
		_mm_store_pd(t,y1);
		for(j-=3;j<je;j++)
		{
			j1    = *(jj++);
			t[0] += *(vv++) * x[j1];
		}
		y[i] = t[0]+t[1];
	}
}

void lis_matvec_crs_p4_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4;
	__m128d		x1,x2;
	__m128d		y1,y2;
	__m128d		v1,v2;
	LIS_SCALAR	xx1,xx2,xx3,xx4;
	LIS_SCALAR	t[2];
	LIS_SCALAR	*vv;


	n    = A->n;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->row[i+1];
		jj = &A->index[js];
		vv = &A->value[js];
		if( je-js<5 )
		{
			t[0] = t[1] = 0;
			j = js;
		}
		else
		{
			y1  = _mm_xor_pd(y1,y1);
			y2  = _mm_xor_pd(y2,y2);
			j1 = *(jj++); xx1 = x[j1];
			j2 = *(jj++); xx2 = x[j2];
			for(j=js+5;j<je;j+=4)
			{
				j3 = *(jj++); xx3 = x[j3];
				j4 = *(jj++); xx4 = x[j4];
				x1  = _mm_set_pd(xx2,xx1);
				v1  = _mm_load_pd(vv);
				x1  = _mm_mul_pd(x1,v1);
				y1  = _mm_add_pd(y1,x1);
				vv += 2;

				j1 = *(jj++); xx1 = x[j1];
				j2 = *(jj++); xx2 = x[j2];
				x2  = _mm_set_pd(xx4,xx3);
				v2  = _mm_load_pd(vv);
				x2  = _mm_mul_pd(x2,v2);
				y2  = _mm_add_pd(y2,x2);
				vv += 2;
			}
			x1  = _mm_set_pd(xx2,xx1);
			v1  = _mm_load_pd(vv);
			x1  = _mm_mul_pd(x1,v1);
			y1  = _mm_add_pd(y1,x1);

			y1  = _mm_add_pd(y1,y2);
			_mm_store_pd(t,y1);
			j -= 3;
		}
		for(;j<je;j++)
		{
			j1    = *(jj++);
			t[0] += *(vv++) * x[j1];
		}
		y[i] = t[0]+t[1];
	}

}
#endif
