/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <math.h>
#include <stdlib.h>
#include "lislib.h"
#include "lis_solver.h"

int lis_qmr(LIS_VECTOR b, LIS_VECTOR x, LIS_VECTOR work[],
		   LIS_SCALAR params[], int options[],
		   LIS_MATRIX A, LIS_PRECON M, LIS_SCALAR status[])
{
	LIS_VECTOR r, rtld, u, p, d, t, t1, q, v;
	LIS_SCALAR tau,rho, rhoold, theta,eta,beta,alpha,w,ww,wold,s,c,etaold,thetaold;
	LIS_SCALAR bnrm2, nrm2;
	LIS_SCALAR tol;
	int iter,maxiter,n,np,m,output;
	int w_dim;
	double times,ptimes;

	n       = A->n;
	np      = A->np + options[LIS_OPTIONS_PADDING]; 
	w_dim   = np;
	maxiter = options[LIS_OPTIONS_MAXITER];
	output  = options[LIS_OPTIONS_OUTPUT];
	tol     = params[LIS_PARAMS_RESID];

	r       = work[0];
	rtld    = work[1];
	u       = work[2];
	p       = work[3];
	d       = work[4];
	t       = work[5];
	t1      = work[6];
	q       = work[7];
	v       = work[8];

	ptimes  = 0.0;

	// Initial Residual
	if( lis_get_initial_residual(A,M,x,b,rtld,r,tol,&bnrm2,status) )
	{
		return LIS_SUCCESS;
	}

	// rtld = r
	// p    = r
	// u    = r
	// d    = 0
	// v    = Ap
	lis_vector_copy(rtld,r);
	lis_vector_copy(r,p);
	lis_vector_copy(r,u);
	lis_vector_set_all(0,d);
	lis_matvec(A,p,t);
	times = lis_wtime();
	lis_psolve(M, t, v);
	ptimes += lis_wtime()-times;

	// rhoold = (r,rtld)
	lis_vector_dot(r,rtld,&rhoold);
	lis_vector_nrm2(r,&tau);
	wold   = tau;
	theta  = 0.0;
	eta    = 0.0;
	thetaold = 0.0;
	etaold = 0.0;

	//	tol = tol * bnrm2;

	iter=1;
	while( iter<=maxiter )
	{
		// alpha = rho / (v,rtld)
		lis_vector_dot(v,rtld,&s);
		/* test breakdown */
		if( s==0.0 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_BREAKDOWN;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			return LIS_BREAKDOWN;
		}
		alpha = rhoold / s;

		// q = u - alpha*v
		// r = r - alpha*A(u + q)
		lis_vector_axpyz(-alpha,v,u,q);
		lis_vector_axpyz(1.0,u,q,t);

		lis_matvec(A,t,t1);
		times = lis_wtime();
		lis_psolve(M, t1, v);
		ptimes += lis_wtime()-times;
		lis_vector_axpy(-alpha,v,r);

		lis_vector_nrm2(r,&w);
		for(m=0;m<2;m++)
//		for(m=2*iter-1;m<=2*iter;m++)
		{
			if( m==0 )
			{
				ww = sqrt(w*wold);
				lis_vector_xpay(u,theta*theta*eta/alpha,d);
			}
			else
			{
				ww = w;
				lis_vector_xpay(q,theta*theta*eta/alpha,d);
			}

			theta = ww / tau; 
			c     = 1.0 / sqrt(1.0 + theta*theta);
			eta   = c * c * alpha;
			tau   = tau * theta * c;
			lis_vector_axpy(eta,d,x);


			nrm2 = tau * sqrt(1.0+m) * bnrm2;
			/* convergence check */
			//nrm2 = lis_vector_nrm20(n,r) * bnrm2;

			if( output )
			{
				if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
				if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
			}

			if( tol >= nrm2 )
			{
				status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_SUCCESS;
				status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
				status[LIS_STATUS_RESID] = nrm2;
				status[LIS_STATUS_PTIMES] = ptimes;
				return LIS_SUCCESS;
			}
		}

		// rho  = (r,rtld)
		// beta = rho / rhoold
		// u    = r + beta*q
		// p    = u + beta*(q + beta*p)
		lis_vector_dot(r,rtld,&rho);
		/* test breakdown */
		if( rho==0.0 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_BREAKDOWN;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			return LIS_BREAKDOWN;
		}
		beta = rho / rhoold;
		lis_vector_axpyz(beta,q,r,u);
		lis_vector_axpy(beta,p,q);
		lis_vector_axpyz(beta,q,u,p);
		lis_matvec(A,p,t1);
		times = lis_wtime();
		lis_psolve(M, t1, v);
		ptimes += lis_wtime()-times;

		rhoold = rho;
		wold   = w;
		iter++;
	}

	status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_MAXITER;
	status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
	status[LIS_STATUS_RESID] = nrm2;
	return LIS_MAXITER;
}

